from ._tissue_segmenter import TissueSegmenter
from ...image._binary_mask import BinaryMask
from ...image._image import Image
from ...image._ubyte_image import UByteImage
from ...process._process import Processor
from ...process.colour._rgb_to_grey import RGBToGrey
import dask.array as da
from skimage.filters import threshold_otsu

class OtsuThresholder(TissueSegmenter):
    def __init__(self, preprocessor: Processor = None, postprocessor: Processor = None, below_thresh: bool = True):
        super().__init__("Otsu Thresholder", preprocessor, postprocessor)
        assert isinstance(below_thresh, bool), "below_thresh must be a boolean"
        self.below_thresh = below_thresh

    def fit(self, image: Image) -> None:
        image = self.preprocessor.run(image)
        if not isinstance(image, UByteImage):
            raise NotImplementedError("OtsuThresholder only supports UByteImage at this time")
        assert image.shape[2] == 1, "Image must be single channel"
        counts, _ = da.histogram(image.image, bins=256, range=[0, 256])
        self.threshold = threshold_otsu(hist=counts.compute(), nbins=256)
        self.fitted = True

    def run(self, image: Image) -> BinaryMask:
        if not self.fitted:
            raise RuntimeError("Model must be fitted before running")
        image = self.preprocessor.run(image)
        if not isinstance(image, UByteImage):
            raise NotImplementedError("OtsuThresholder only supports UByteImage at this time")
        assert image.shape[2] == 1, "Image must be single channel"
        img = image.image
        img = da.squeeze(img, axis=2)
        if self.below_thresh:
            mask = img < self.threshold
        else:
            mask = img > self.threshold
        mask = BinaryMask(mask, image.info, "TissueMask")
        mask = self.postprocessor.run(mask)
        return mask

    def fit_and_run(self, image: Image) -> Image:
        self.fit(image)
        return self.run(image)

    def _default_preprocessor(self) -> Processor:
        proc = Processor()
        proc.add_process(RGBToGrey())
        return proc

    def _default_postprocessor(self) -> Processor:
        return Processor()


from .._process import Process
from ...image._image import Image
from ...image._ubyte_image import UByteImage

class GammaContrast(Process):
    def __init__(self, gamma: float = 1.0) -> None:
        super().__init__(f"Contrast adjustment using gamma correction of {str(gamma)}")
        assert isinstance(gamma, float) or isinstance(gamma, int), "gamma must be convertable to float."
        self.gamma = gamma

    def run(self, image: Image) -> Image:
        assert isinstance(image, Image), "image must be of type Image."
        if not isinstance(image, UByteImage):
            raise NotImplementedError("GammaContrast only supports UByteImage at this time. Please raise an issue to change this.")
        img = image.image
        gamma = self.gamma
        if isinstance(image, UByteImage):
            img = img/255.0
        img = img ** gamma
        if isinstance(image, UByteImage):
            img = img*255.0
            img = img.astype('uint8')
            return UByteImage(img, image.info, image.channel_names)
        else:
            raise NotImplementedError("GammaContrast only supports UByteImage at this time. Please raise an issue to change this.")




from scematk.process._process import Processor
from .._segmenter import Segmenter

class TissueSegmenter(Segmenter):
    def __init__(self, name: str, preprocessor: Processor = None, postprocessor: Processor = None):
        super().__init__(name, preprocessor, postprocessor)


from .._process import Process
from ...image._image import Image
from ...image._ubyte_image import UByteImage
import dask.array as da

class LinearContrast(Process):
    def __init__(self, slope: float = 1.0, intercept: float = 0.0) -> None:
        super().__init__(f"Contrast adjustment using linear correction with a slope of {str(slope)} and an intercept of {str(intercept)}")
        assert isinstance(slope, float) or isinstance(slope, int), "slope must be convertable to float."
        assert isinstance(intercept, float) or isinstance(intercept, int), "intercept must be convertable to float."
        self.slope = slope
        self.intercept = intercept

    def run(self, image: Image) -> Image:
        assert isinstance(image, Image), "image must be of type Image."
        if not isinstance(image, UByteImage):
            raise NotImplementedError("GammaContrast only supports UByteImage at this time. Please raise an issue to change this.")
        img = image.image
        slope = self.slope
        intercept = self.intercept
        if isinstance(image, UByteImage):
            img = img/255.0
        img = img * slope + intercept
        img = da.clip(img, 0, 1)
        if isinstance(image, UByteImage):
            img = img*255.0
            img = img.astype('uint8')
            return UByteImage(img, image.info, image.channel_names)
        else:
            raise NotImplementedError("GammaContrast only supports UByteImage at this time. Please raise an issue to change this.")


from ._eddie import eddie_profile

def get_profile(profile: str) -> dict:
    assert isinstance(profile, str), "Profile must be a string."
    if profile == "eddie":
        return eddie_profile.copy()
    else:
        raise ValueError(f"Profile {profile} not found.")


from ...process import Process
from ...image._image import Image
from ...image._ubyte_image import UByteImage
from dask_image.ndfilters import maximum_filter, median_filter, minimum_filter, percentile_filter
from skimage.morphology import disk
import numpy as np

class MaximumFilter(Process):
    def __init__(self, radius: float, metric: str = 'micron') -> None:
        assert isinstance(radius, (int, float)), 'radius must be a number'
        assert radius > 0, 'radius must be positive'
        assert isinstance(metric, str), 'metric must be a string'
        assert metric in ['micron', 'pixel'], 'metric must be either "micron" or "pixel"'
        self.radius = radius
        self.metric = metric
        super().__init__(name=f'Maximum Filter with a radius of {radius} {metric}s')

    def run(self, image: Image) -> Image:
        assert isinstance(image, Image), 'image must be an Image'
        radius = image.pixel_from_micron(self.radius) if self.metric == 'micron' else self.radius
        footprint = disk(radius)
        footprint = np.expand_dims(footprint, axis=-1)
        img = image.image
        img = maximum_filter(img, footprint=footprint)
        if isinstance(image, UByteImage):
            img = img.astype('uint8')
            return UByteImage(img, image.info, image.channel_names)
        else:
            raise NotImplementedError('MaximumFilter only supports UByteImage objects at this time')

class MedianFilter(Process):
    def __init__(self, radius: float, metric: str = 'micron') -> None:
        assert isinstance(radius, (int, float)), 'radius must be a number'
        assert radius > 0, 'radius must be positive'
        assert isinstance(metric, str), 'metric must be a string'
        assert metric in ['micron', 'pixel'], 'metric must be either "micron" or "pixel"'
        self.radius = radius
        self.metric = metric
        super().__init__(name=f'Median Filter with a radius of {radius} {metric}s')

    def run(self, image: Image) -> Image:
        assert isinstance(image, Image), 'image must be an Image'
        radius = image.pixel_from_micron(self.radius) if self.metric == 'micron' else self.radius
        footprint = disk(radius)
        footprint = np.expand_dims(footprint, axis=-1)
        img = image.image
        img = median_filter(img, footprint=footprint)
        if isinstance(image, UByteImage):
            img = img.astype('uint8')
            return UByteImage(img, image.info, image.channel_names)
        else:
            raise NotImplementedError('MedianFilter only supports UByteImage objects at this time')

class MinimumFilter(Process):
    def __init__(self, radius: float, metric: str = 'micron') -> None:
        assert isinstance(radius, (int, float)), 'radius must be a number'
        assert radius > 0, 'radius must be positive'
        assert isinstance(metric, str), 'metric must be a string'
        assert metric in ['micron', 'pixel'], 'metric must be either "micron" or "pixel"'
        self.radius = radius
        self.metric = metric
        super().__init__(name=f'Minimum Filter with a radius of {radius} {metric}s')

    def run(self, image: Image) -> Image:
        assert isinstance(image, Image), 'image must be an Image'
        radius = image.pixel_from_micron(self.radius) if self.metric == 'micron' else self.radius
        footprint = disk(radius)
        footprint = np.expand_dims(footprint, axis=-1)
        img = image.image
        img = minimum_filter(img, footprint=footprint)
        if isinstance(image, UByteImage):
            img = img.astype('uint8')
            return UByteImage(img, image.info, image.channel_names)
        else:
            raise NotImplementedError('MinimumFilter only supports UByteImage objects at this time')

class PercentileFilter(Process):
    def __init__(self, radius: float, percentile: float, metric: str = 'micron') -> None:
        assert isinstance(radius, (int, float)), 'radius must be a number'
        assert radius > 0, 'radius must be positive'
        assert isinstance(percentile, (int, float)), 'percentile must be a number'
        assert 0 <= percentile <= 100, 'percentile must be between 0 and 100'
        assert isinstance(metric, str), 'metric must be a string'
        assert metric in ['micron', 'pixel'], 'metric must be either "micron" or "pixel"'
        self.radius = radius
        self.percentile = percentile
        self.metric = metric
        super().__init__(name=f'Percentile Filter with a radius of {radius} {metric}s and a percentile of {percentile}')

    def run(self, image: Image) -> Image:
        assert isinstance(image, Image), 'image must be an Image'
        radius = image.pixel_from_micron(self.radius) if self.metric == 'micron' else self.radius
        footprint = disk(radius)
        footprint = np.expand_dims(footprint, axis=-1)
        img = image.image
        img = percentile_filter(img, footprint=footprint, percentile=self.percentile)
        if isinstance(image, UByteImage):
            img = img.astype('uint8')
            return UByteImage(img, image.info, image.channel_names)
        else:
            raise NotImplementedError('PercentileFilter only supports UByteImage objects at this time')


from .._process import Process
from ...image._image import Image
from ...image._ubyte_image import UByteImage
import dask.array as da

class MinMaxContrast(Process):
    def __init__(self) -> None:
        super().__init__("Perform min-max contrast enhancement per channel")

    def run(self, image: Image) -> Image:
        img = image.image
        if isinstance(image, UByteImage):
            img = img / 255
        minc = da.min(img, axis=(0, 1))
        img = img - minc
        maxc = da.max(img, axis=(0, 1))
        img = img / maxc
        if isinstance(image, UByteImage):
            img = img * 255
            img = img.astype("uint8")
            return UByteImage(img, image.info, image.channel_names)
        else:
            raise ValueError("MinMaxContrast can only be applied to UByteImage")


from ..image._image import Image
from ..process._process import Processor
from abc import ABC, abstractmethod

class Model(ABC):
    def __init__(self, name: str, preprocessor: Processor = None, postprocessor: Processor = None) -> None:
        preprocessor = preprocessor if preprocessor is not None else self._default_preprocessor()
        postprocessor = postprocessor if postprocessor is not None else self._default_postprocessor()
        assert isinstance(name, str), "name must be a string"
        assert isinstance(preprocessor, Processor), "preprocessor must be a Processor instance"
        assert isinstance(postprocessor, Processor), "postprocessor must be a Processor instance"
        self.name = name
        self.preprocessor = preprocessor
        self.postprocessor = postprocessor
        self.fitted = False

    @abstractmethod
    def fit(self, image: Image) -> None:
        pass

    @abstractmethod
    def run(self, image: Image) -> Image:
        pass

    @abstractmethod
    def fit_and_run(self, image: Image) -> Image:
        pass

    @abstractmethod
    def _default_preprocessor(self) -> Processor:
        pass

    @abstractmethod
    def _default_postprocessor(self) -> Processor:
        pass


from .._abc._model import Model
from ..process._process import Processor

class Segmenter(Model):
    def __init__(self, name: str, preprocessor: Processor = None, postprocessor: Processor = None):
        super().__init__(name, preprocessor, postprocessor)


from ...image._binary_mask import BinaryMask
from ...process._process import Process
from dask_image.ndmorph import binary_closing
from skimage.morphology import disk


class BinaryClosing(Process):
    def __init__(self, radius: float, metric: str = 'micron') -> None:
        assert isinstance(radius, (int, float)), 'radius must be a number'
        assert radius > 0, 'radius must be positive'
        assert isinstance(metric, str), 'metric must be a string'
        assert metric in ['micron', 'pixel'], 'metric must be either "micron" or "pixel"'
        self.radius = radius
        self.metric = metric
        super().__init__(name=f'Binary Closing with a radius of {radius} {metric}s')

    def run(self, image: BinaryMask) -> BinaryMask:
        assert isinstance(image, BinaryMask), 'image must be a BinaryMask'
        radius = image.pixel_from_micron(self.radius) if self.metric == 'micron' else self.radius
        mask = image.image
        footprint = disk(radius)
        mask = binary_closing(mask, structure=footprint)
        return BinaryMask(mask, image.info, image.channel_names)



from .profiles._get_profile import get_profile
from dask_jobqueue import HTCondorCluster, LSFCluster, MoabCluster, OARCluster, PBSCluster, SGECluster, SLURMCluster

def quick_connect(profile: str, minimum: int, maximum: int, min_n_wait: int = None, timeout: int = 300):
    assert isinstance(profile, str), "profile must be a string"
    assert isinstance(minimum, int), "minimum must be an integer"
    assert isinstance(maximum, int), "maximum must be an integer"
    if min_n_wait is None:
        min_n_wait = minimum
    assert isinstance(min_n_wait, int), "min_n_wait must be an integer"
    assert min_n_wait <= minimum, "min_n_wait must be less than or equal to minimum"
    assert isinstance(timeout, int), "timeout must be an integer"
    profile_dict = get_profile(profile)
    cluster_type = profile_dict.pop('cluster_type')
    if cluster_type == 'HTCondor':
        cluster = HTCondorCluster(**profile_dict)
    elif cluster_type == 'LSF':
        cluster = LSFCluster(**profile_dict)
    elif cluster_type == 'Moab':
        cluster = MoabCluster(**profile_dict)
    elif cluster_type == 'OAR':
        cluster = OARCluster(**profile_dict)
    elif cluster_type == 'PBSCluster':
        cluster = PBSCluster(**profile_dict)
    elif cluster_type == 'SGE':
        cluster = SGECluster(**profile_dict)
    elif cluster_type == 'SLURM':
        cluster = SLURMCluster(**profile_dict)
    else:
        raise ValueError(f"Unsupported cluster type: {cluster_type}")
    try:
        cluster.adapt(minimum=minimum, maximum=maximum)
    except Exception as e:
        cluster.close()
        raise e
    try:
        cluster.wait_for_workers(min_n_wait, timeout=timeout)
    except TimeoutError:
        cluster.close()
        raise TimeoutError(f"Timeout: {min_n_wait} workers did not start within {timeout} seconds")
    except Exception as e:
        cluster.close()
        raise e
    return cluster